@@ -15,9 +15,12 @@ from scipy import spatial
 to_delete = re.compile(
     r'[」「@#$%^&»«“”„‟"\x1a⓪①②③④⑤⑥⑦⑧⑨⑩⑴⑵⑶⑷⑸⑹⑺⑻⑼⑽*\(\)\[\]\n\/\-\:•＂＃＄％＆＇（）＊＋－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､、〃》【】〔〕〖〗〘〙〜〟〰〾〿–—‘’‛‧﹏〉]+')
 
-def get_line_vectors(lines, model_name, embed_batch_size=10, normalize_embeddings=True, show_progress_bar=False):
+def get_line_vectors(lines, model_name, embed_batch_size=10, normalize_embeddings=True, show_progress_bar=False, model=None):
     """Calculate embedding of the string"""
-    return model_dispatcher.models[model_name].embed(lines, embed_batch_size, normalize_embeddings, show_progress_bar)
+    if model:
+        return model.encode(lines, batch_size=embed_batch_size, normalize_embeddings=normalize_embeddings, show_progress_bar=show_progress_bar)
+    else:
+        return model_dispatcher.models[model_name].embed(lines, embed_batch_size, normalize_embeddings, show_progress_bar)
 
 
 def clean_lines(lines):
@@ -25,7 +28,27 @@ def clean_lines(lines):
     return [re.sub(to_delete, '', line) for line in lines]
 
 
-def process_batch(lines_from_batch, lines_to_batch, line_ids_from, line_ids_to, batch_number, model_name, window, embed_batch_size, normalize_embeddings, show_progress_bar, save_pic=False, lang_name_from="", lang_name_to="", img_path="", show_info=False, show_regression=False):
+def process_batch(lines_from_batch,
+                    lines_to_batch,
+                    proxy_from_batch,
+                    proxy_to_batch,
+                    line_ids_from,
+                    line_ids_to,
+                    batch_number,
+                    model_name,
+                    window,
+                    embed_batch_size,
+                    normalize_embeddings,
+                    show_progress_bar,
+                    save_pic=False,
+                    lang_name_from="",
+                    lang_name_to="",
+                    img_path="",
+                    show_info=False,
+                    show_regression=False,
+                    model=None,
+                    use_proxy_from=False,
+                    use_proxy_to=False):
     """Do the actual alignment process logic"""
     # try:
     logging.info(f"Batch {batch_number}. Calculating vectors.")
@@ -33,8 +56,15 @@ def process_batch(lines_from_batch, lines_to_batch, line_ids_from, line_ids_to,
     # vectors1 = [*get_line_vectors(clean_lines(lines_from_batch), model_name, embed_batch_size, normalize_embeddings, show_progress_bar)]
     # vectors2 = [*get_line_vectors(clean_lines(lines_to_batch), model_name, embed_batch_size, normalize_embeddings, show_progress_bar)]
 
-    vectors1 = [*get_line_vectors(lines_from_batch, model_name, embed_batch_size, normalize_embeddings, show_progress_bar)]
-    vectors2 = [*get_line_vectors(lines_to_batch, model_name, embed_batch_size, normalize_embeddings, show_progress_bar)]
+    if use_proxy_from:
+        vectors1 = [*get_line_vectors(proxy_from_batch, model_name, embed_batch_size, normalize_embeddings, show_progress_bar, model)]
+    else:
+        vectors1 = [*get_line_vectors(lines_from_batch, model_name, embed_batch_size, normalize_embeddings, show_progress_bar, model)]
+
+    if use_proxy_to:
+        vectors2 = [*get_line_vectors(proxy_to_batch, model_name, embed_batch_size, normalize_embeddings, show_progress_bar, model)]
+    else:
+        vectors2 = [*get_line_vectors(lines_to_batch, model_name, embed_batch_size, normalize_embeddings, show_progress_bar, model)]
 
     logging.debug(
         f"Batch {batch_number}. Vectors calculated. len(vectors1)={len(vectors1)}. len(vectors2)={len(vectors2)}.")
@@ -74,15 +104,21 @@ def process_batch(lines_from_batch, lines_to_batch, line_ids_from, line_ids_to,
     #     return [], []
 
 
-def align_texts(splitted_from, splitted_to, model_name, batch_size, window, batch_ids=[], save_pic=False, lang_from="", lang_to="", img_path="", embed_batch_size=10, normalize_embeddings=True, show_progress_bar=False, shift=0, show_info=False, show_regression=False):
+def align_texts(splitted_from, splitted_to, model_name, batch_size, window, batch_ids=[], save_pic=False, lang_from="", lang_to="", img_path="", embed_batch_size=10, normalize_embeddings=True, show_progress_bar=False, shift=0, show_info=False, show_regression=False, proxy_from=[], proxy_to=[]):
     result = []
-    task_list = [(lines_from_batch, lines_to_batch, line_ids_from, line_ids_to, batch_id)
-                 for lines_from_batch, lines_to_batch,
-                 line_ids_from, line_ids_to, batch_id
-                 in get_batch_intersected(splitted_from, splitted_to, batch_size, window, batch_ids, batch_shift=shift)]
+    task_list = [(lines_from_batch, lines_to_batch, proxy_from_batch, proxy_to_batch, line_ids_from, line_ids_to, batch_id)
+                 for
+                    lines_from_batch,
+                    lines_to_batch,
+                    proxy_from_batch,
+                    proxy_to_batch,
+                    line_ids_from,
+                    line_ids_to,
+                    batch_id
+                 in get_batch_intersected(splitted_from, splitted_to, batch_size, window, batch_ids, batch_shift=shift, iter3=proxy_from, iter4=proxy_to)]
 
-    for lines_from_batch, lines_to_batch, line_ids_from, line_ids_to, batch_id in task_list:
-        texts_from, texts_to = process_batch(lines_from_batch, lines_to_batch, line_ids_from,
+    for lines_from_batch, lines_to_batch, proxy_from_batch, proxy_to_batch, line_ids_from, line_ids_to, batch_id in task_list:
+        texts_from, texts_to = process_batch(lines_from_batch, lines_to_batch, proxy_from_batch, proxy_to_batch, line_ids_from,
                                              line_ids_to, batch_id, model_name, window, embed_batch_size, normalize_embeddings, show_progress_bar, save_pic, lang_from, lang_to, img_path, show_info=show_info, show_regression=show_regression)
         result.append((batch_id, texts_from, texts_to))
 
@@ -92,20 +128,26 @@ def align_texts(splitted_from, splitted_to, model_name, batch_size, window, batc
     return result
 
 
-def align_db(db_path, model_name, batch_size, window, batch_ids=[], save_pic=False, lang_from="", lang_to="", img_path="", embed_batch_size=10, normalize_embeddings=True, show_progress_bar=False, shift=0, show_info=False, show_regression=False):
+def align_db(db_path, model_name, batch_size, window, batch_ids=[], save_pic=False, lang_from="", lang_to="", img_path="", embed_batch_size=10, normalize_embeddings=True, show_progress_bar=False, shift=0, show_info=False, show_regression=False, model=None, proxy_from=[], proxy_to=[]):
     result = []
     splitted_from = get_splitted_from(db_path)
     splitted_to = get_splitted_to(db_path)
-    task_list = [(lines_from_batch, lines_to_batch, line_ids_from, line_ids_to, batch_id)
-                 for lines_from_batch, lines_to_batch,
-                 line_ids_from, line_ids_to, batch_id
-                 in get_batch_intersected(splitted_from, splitted_to, batch_size, window, batch_ids, batch_shift=shift)]
+    task_list = [(lines_from_batch, lines_to_batch, proxy_from_batch, proxy_to_batch, line_ids_from, line_ids_to, batch_id)
+                 for
+                    lines_from_batch,
+                    lines_to_batch,
+                    proxy_from_batch,
+                    proxy_to_batch,
+                    line_ids_from,
+                    line_ids_to,
+                    batch_id
+                 in get_batch_intersected(splitted_from, splitted_to, batch_size, window, batch_ids, batch_shift=shift, iter3=proxy_from, iter4=proxy_to)]
 
     count = 0
     for lines_from_batch, lines_to_batch, line_ids_from, line_ids_to, batch_id in task_list:
         print("batch:", count)
         texts_from, texts_to = process_batch(lines_from_batch, lines_to_batch, line_ids_from,
-                                             line_ids_to, batch_id, model_name, window, embed_batch_size, normalize_embeddings, show_progress_bar, save_pic, lang_from, lang_to, img_path, show_info=show_info, show_regression=show_regression)
+                                             line_ids_to, batch_id, model_name, window, embed_batch_size, normalize_embeddings, show_progress_bar, save_pic, lang_from, lang_to, img_path, show_info=show_info, show_regression=show_regression, model=model, proxy_from=[], proxy_to=[])
         result.append((batch_id, texts_from, texts_to, shift, window))
         count += 1
 
@@ -136,6 +178,22 @@ def get_splitted_to(db_path):
     return [x[0] for x in res]
 
 
+def get_proxy_from(db_path):
+    """Get lines from proxy_from"""
+    with sqlite3.connect(db_path) as db:
+        res = db.execute(
+            f'select f.proxy_text from splitted_from f order by f.id').fetchall()
+    return [x[0] for x in res]
+
+
+def get_proxy_to(db_path):
+    """Get lines from proxy_to"""
+    with sqlite3.connect(db_path) as db:
+        res = db.execute(
+            f'select t.proxy_text from splitted_to t order by t.id').fetchall()
+    return [x[0] for x in res]
+
+
 def best_per_row_with_ones(sim_matrix):
     """Transfor matrix by leaving only best match"""
     sim_matrix_best = np.zeros_like(sim_matrix)
@@ -144,13 +202,18 @@ def best_per_row_with_ones(sim_matrix):
     return sim_matrix_best
 
 
-def get_batch_intersected(iter1, iter2, n, window, batch_ids=[], batch_shift=0):
+def get_batch_intersected(iter1, iter2, n, window, batch_ids=[], batch_shift=0, iter3=[], iter4=[]):
     """Get batch with an additional window"""
     l1 = len(iter1)
     l2 = len(iter2)
     k = int(round(n * l2/l1))
     kdx = 0 - k
 
+    if not iter3:
+        iter3 = ['' for _ in range(l1)]
+    if not iter4:
+        iter4 = ['' for _ in range(l2)]
+
     if k < window*2:
         # subbatches will be intersected
         logging.warning(
@@ -162,6 +225,8 @@ def get_batch_intersected(iter1, iter2, n, window, batch_ids=[], batch_shift=0):
         if counter in batch_ids or len(batch_ids) == 0:
             yield iter1[ndx:min(ndx + n, l1)], \
                 iter2[max(0, kdx - window + batch_shift):min(kdx + k + window + batch_shift, l2)], \
+                iter3[ndx:min(ndx + n, l1)], \
+                iter4[max(0, kdx - window + batch_shift):min(kdx + k + window + batch_shift, l2)], \
                 list(range(ndx, min(ndx + n, l1))), \
                 list(range(max(0, kdx - window + batch_shift), min(kdx + k + window + batch_shift, l2))), \
                 counter
@@ -271,7 +336,7 @@ def init_document_db(db_path):
         db.execute(
             'create table history(id integer primary key, operation text, batch_id integer, insert_ts text, parameters text)')
         db.execute(
-            'create table meta(id integer primary key, key text, val text, occurence integer, par_id integer)')
+            'create table meta(id integer primary key, key text, val text, occurence integer, par_id integer, deleted integer DEFAULT 0, comment text DEFAULT "")')
         db.execute(
             'create table languages(id integer primary key, key text, val text)')
         db.execute(
@@ -279,19 +344,19 @@ def init_document_db(db_path):
         db.execute('insert into version(version) values (?)', (con.DB_VERSION,))
 
 
-def fill_db_from_files(db_path, lang_from, lang_to, splitted_from, splitted_to, proxy_from, proxy_to):
+def fill_db_from_files(db_path, lang_from, lang_to, splitted_from_path, splitted_to_path, proxy_from_path, proxy_to_path):
     """Fill document database (alignment) with prepared document lines"""
     if not os.path.isfile(db_path):
         logging.info(f"Initializing database {db_path}")
         init_document_db(db_path)
     lines = []
-    if os.path.isfile(splitted_from):
-        with open(splitted_from, mode="r", encoding="utf-8") as input_path:
+    if os.path.isfile(splitted_from_path):
+        with open(splitted_from_path, mode="r", encoding="utf-8") as input_path:
             lines = input_path.readlines()
         lines, meta, meta_par_ids = handle_marks(lines)
         lines_proxy = []
-        if os.path.isfile(proxy_from):
-            with open(proxy_from, mode="r", encoding="utf-8") as input_path:
+        if os.path.isfile(proxy_from_path):
+            with open(proxy_from_path, mode="r", encoding="utf-8") as input_path:
                 lines_proxy = input_path.readlines()
         if len(lines) == len(lines_proxy):
             data = zip(lines, lines_proxy)
@@ -301,13 +366,13 @@ def fill_db_from_files(db_path, lang_from, lang_to, splitted_from, splitted_to,
             db.executemany("insert into splitted_from(text, proxy_text, exclude, paragraph, h1, h2, h3, h4, h5, divider) values (?,?,?,?,?,?,?,?,?,?)", [
                            (text[0].strip(), proxy.strip(), 0, text[1][0], text[1][1], text[1][2], text[1][3], text[1][4], text[1][5], text[1][6]) for text, proxy in data])
             db.executemany("insert into meta(key, val, occurence, par_id) values(?,?,?,?)", flatten_meta(meta, meta_par_ids,"from"))
-    if os.path.isfile(splitted_to):
-        with open(splitted_to, mode="r", encoding="utf-8") as input_path:
+    if os.path.isfile(splitted_to_path):
+        with open(splitted_to_path, mode="r", encoding="utf-8") as input_path:
             lines = input_path.readlines()
         lines, meta, meta_par_ids = handle_marks(lines)
         lines_proxy = []
-        if os.path.isfile(proxy_to):
-            with open(proxy_to, mode="r", encoding="utf-8") as input_path:
+        if os.path.isfile(proxy_to_path):
+            with open(proxy_to_path, mode="r", encoding="utf-8") as input_path:
                 lines_proxy = input_path.readlines()
         if len(lines) == len(lines_proxy):
             data = zip(lines, lines_proxy)
@@ -404,7 +469,7 @@ def handle_marks(lines):
     for line in lines:
         next_par = False
         line = line.strip()
-        
+
         if line.endswith(p_ending):
             #remove last occurence of PARAGRAPH_MARK
             line = ''.join(line.rsplit(preprocessor.PARAGRAPH_MARK, 1))
